/* include gnu.gpl, (*) 2000 Micha 'malekith' Moskal <malekith@topnet.pl>. 
 * $Id: input.c,v 1.6 2000/08/25 20:04:08 malekith Exp $ */

#include "bot.h"
#include "proc.h"
#include "var.h"
#include "script_ext.h"

/* aliases syntax:
anything .... == /anything
//anything -> don't resolve anymore

magic % : `%' follwed by X is expanded. X is one of :
 
%        	`%' sign itself, so %% gives you one %
0 ... 99	positional parameters of alias, %0 is alias name, %1 is the
		first argument and so on.
0+ ... 99+	%0+ is entire alias, %1+ are all arguments from the first,
		%2+ -- all from the second and so on
!1 ... !99	positonal arguments, which are required
!1+ ... !99+	same, plus the rest
{default}1 ...	arguments, which if aren't present gets the default value
... {blurb}99
c		name of actual channel
n		nick of the user

additionaly if no %1 etc. arguments were used, alias behave as if
it was follwed by a %1+, doing simply
/alias ml /msg lynn_
is same as
/alias ml /msg lynn_ %1+
and does what you expect.
*/

struct command {
	struct command *next;
	char *name;
	
	int alias;
	char *code;
};

struct input_line {
	const char *win;
	int argc;
	char **argv;
	char **argv_plus;
	char *args;
	char *free_me;
	
	int error;
};

struct alias_ctx {
	const char *typeahead;
	char *free_me;
	char *p;
	struct input_line *i;
	int error;
};

static hash *aliases, *commands;

static var *c_add_command(struct frame *f)
{
	char *name = getsarg(f, "name");
	char *code = getsarg(f, "code");
	struct command *n;

	hash_del(commands, name);
	n = NEW(struct command);
	n->name = xstrdup(name);
	n->alias = 0;
	n->code = xstrdup(code);
	hash_add(commands, (var*)n);

	return 0;
}

static var *c_add_alias(struct frame *f)
{
	char *name = getsarg(f, "name");
	char *code = getsarg(f, "code");
	struct command *n;

	hash_del(aliases, name);
	n = NEW(struct command);
	n->name = xstrdup(name);
	n->alias = 1;
	n->code = xstrdup(code);
	hash_add(aliases, (var*)n);
	return 0;
}

static var *c_del_alias(struct frame *f)
{
	char *name = getsarg(f, "name");

	hash_del(aliases, name);
	return 0;
}

static char *get_arg(char **pp, char **plus, int real)
{
	char *p, *r;
	int stop = 0;
	
	if (plus)
		*plus = 0;
		
	if ((p = *pp) == 0)
		return 0;
	
	if (plus)
		*plus = xstrdup(p);
		
	while (*p == ' ')
		p++;
	if (*p == '\'' || *p == '"')
		stop = *p++;
	else
		stop = ' ';
	r = p;
	
	while (*p && *p != stop)
		p++;
		
	if (*p) {
		if (real)
			*p = 0;
		p++;
	}
	if (*p)
		*pp = p;
	else
		*pp = 0;

	return r;
}

static void split_args(struct input_line *i, int need_plus)
{
	char *p, **op, **o;
	
	i->argc = 1;
	p = i->args;
	
	while (get_arg(&p, 0, 0))
		i->argc++;

	if (need_plus)
		op = i->argv_plus = (char**)xmalloc((i->argc + 1) * sizeof(char*));
	else
		i->argv_plus = 0;
	o = i->argv = (char**)xmalloc((i->argc + 1) * sizeof(char*));
	
	p = i->args;
	
	while ((*o++ = get_arg(&p, need_plus ? op++ : 0, 1)))
		/* nothing */ ;
}

static void kill_input_line(struct input_line *i)
{
	char **o = i->argv_plus;
	
	if (o)
		while (*o)
			xfree(*o++);
		
	xfree(i->argv_plus);
	xfree(i->argv);
	i->argv = i->argv_plus = 0;
	xfree(i->free_me);
	i->free_me = 0;
	i->error = 0;
	i->argc = 0;
	i->args = 0;
}

static void set_typeahead(struct alias_ctx *ctx)
{
	char *p, *def = "";
	int i;
	
	p = ctx->p;
	
	switch (*p) {
	case '%':
		ctx->typeahead = "%";
		p++;
		break;
	case 'n':
		ctx->typeahead = sv("nick")->var_string;
		p++;
		break;
	case 'c':
		ctx->typeahead = ctx->i->win;
		p++;
		break;
	case '{':
		p++;
		for (i = 0; p[i]; i++)
			if (p[i] == '}')
				break;
		def = ctx->free_me = xstrndup(p, i);
		p += i;
		if (*p)
			p++;
		goto defl;
	case '!':
		p++;
		def = 0;
		/* fall through */
	default:
	defl:
		if (*p >= '0' && *p <= '9') {
			i = *p++ - '0';
			if (i && *p >= '0' && *p <= '9') {
				i *= 10;
				i += *p++ - '0';
			} 

			if (ctx->i->argc <= i) {
				if (def)
					ctx->typeahead = def;
				else
					ctx->error = -i;
				if (*p == '+')
					p++;
			} else {
				if (*p == '+') {
					p++;
					ctx->typeahead = ctx->i->argv_plus[i];
				} else
					ctx->typeahead = ctx->i->argv[i];
			}
		} else {
			if (ctx->free_me) { 
				xfree(ctx->free_me);
				ctx->free_me = 0;
			}
			ctx->error = ((unsigned char)*p) + 1;
		}
	}
	ctx->p = p;
}

static int get_alias_char(struct alias_ctx *ctx)
{
	int r = 0;

	if (ctx->typeahead == 0 || *ctx->typeahead == 0) {
		if (ctx->free_me) {
			xfree(ctx->free_me);
			ctx->free_me = 0;
		}
		do {
			ctx->typeahead = 0;
			if ((r = *ctx->p++) == '%')
				set_typeahead(ctx);
			else
				break;
			if (ctx->error)
				return 0;
		} while (ctx->typeahead == 0 || *ctx->typeahead == 0);
	}

	if (ctx->typeahead) {
		r = *ctx->typeahead++;
		if (*ctx->typeahead == 0) {
			ctx->typeahead = 0;
			xfree(ctx->free_me);
			ctx->free_me = 0;
		}
	}

	return r;
}

static void unalias(struct input_line *l, int lev)
{
	struct alias_ctx ctx;
	struct command *v;
	char *r, *p;
	int len, q = 0;

	if (lev > 16) {
		prt(l->win, "\e<user_error> infinite alias loop");
		l->error = 1;
		return;
	}

	v = (struct command*)hash_find_ex(aliases, l->argv[0]);
	
	if (v == 0)
		return;	/* not an alias */
	memset(&ctx, 0, sizeof(ctx));
	
	ctx.i = l;
	ctx.p = v->code;
	len = 0;
	while (get_alias_char(&ctx))
		len++;

	if (ctx.error) {
		if (ctx.error > 0) 
			prtf(l->win, "\e<user_error> unknown alias spell %%%c", 
				ctx.error - 1);
		else
			prtf(l->win, "\e<user_error> required argument %%%d "
				"for alias missing", -ctx.error);
		l->error = 2;
		return;
	}

	r = p = (char*)xmalloc(len + 1);
	memset(&ctx, 0, sizeof(ctx));
	ctx.i = l;
	ctx.p = v->code;
	while ((*p++ = get_alias_char(&ctx)))
		/* nothing */;

	kill_input_line(l);
	l->free_me = r;

	if (*r == '/')
		r++;
	if (*r == '/') {
		q = 1;
		r++;
	}
	
	l->args = r;
	split_args(l, 1);
	if (!q)
		unalias(l, ++lev);
}

static void exec_command(struct input_line *i)
{
	struct command *c;
	int z;
	char *k, **op, **o;

	c = (struct command *) hash_find_ex(commands, i->argv[0]);
	if (c == 0) {
		prtf(i->win, "\e<user_error> no, no, bad magic word: %s",
		    i->argv[0]);
		return;
	}
	
	set_iv(hv("cmd", "argc"), i->argc);
	set_sv(hv("cmd", "win"), i->win);
	
	/* current window */
	set_sv(sv("window"), i->win);
	
	o = i->argv;
	op = i->argv_plus;
	for (z = 0; z < 20; z++) {
		k = xf("argv%dp", z);
		set_sv(hv("cmd", k), *op ? *op++ : "");
		*(strchr(k, 0) - 1) = 0;	/* delete trailing `p' */
		set_sv(hv("cmd", k), *o ? *o++ : "");
		xfree(k);
	}
	
	eval(c->code);
}

void do_input(const char *win, const char *x)
{
	char *line, *p;
	struct input_line l;
	int q = 0;

	if (win == 0 || *win == 0 || strcasecmp(win, "status") == 0)
		win = "status";
	
	l.free_me = line = xstrdup(x);

	p = line;
	memset(&l, 0, sizeof(l));
	l.win = win;
	/* reconize commands after any number of blanks,
	 * nice if you just wrote ``  /kick sucker'' and 
	 * press enter ... try to do this in BX ;) */
	while (*p == ' ')
		p++;
	if (*p == '/') {
		p++;
		if (*p == '/') {
			p++;
			q = 1;
		}
		l.args = p;
		split_args(&l, 1);
		if (q == 0)
			unalias(&l, 0);
	} else {
		l.argv = (char**)xmalloc(3 * sizeof(char*));
		l.argv_plus = (char**)xmalloc(3 * sizeof(char*));
		l.argv[0] = "_msg";
		l.argv[1] = line;
		l.argv[2] = 0;
		l.argv_plus[0] = xf("_msg %s", line);
		l.argv_plus[1] = xstrdup(line);
		l.argv_plus[2] = 0;
		l.args = 0;
	}
	
	if (!l.error)
		exec_command(&l);

	kill_input_line(&l);
}

static void free_command(var *p)
{
	struct command *c;
	c = (struct command *)p;
	xfree(c->name);
	xfree(c->code);
	xfree(c);
}

void init_input()
{
	aliases = hash_new(0);
	commands = hash_new(0);
	aliases->free_entry = free_command;
	commands->free_entry = free_command;
	add_cproc("add_command", "string name, string code", c_add_command);
	add_cproc("add_alias", "string name, string code", c_add_alias);
	add_cproc("del_alias", "string name", c_del_alias);
}
