/* include gnu.gpl, (*) 2000 Micha 'malekith' Moskal <malekith@topnet.pl>. 
 * $Id: socklist.c,v 1.8 2000/08/25 22:01:21 malekith Exp $ */

#include "bot.h"
#include "var.h"
#include "script_ext.h"
#include <unistd.h>
#include <netinet/in.h>
#include <sys/socket.h>

struct socket *sock_head;
static int client_handler(int fd, void *data);

void xclose(int fd)
{
	struct socket *p, *tmp;
	close(fd);
	if (sock_head->fd == fd) {
		p = sock_head;
		sock_head = sock_head->next;
		xfree(p);
		return;
	}
	for (p = sock_head; p && p->next; p = p->next)
		if (p->next->fd == fd) {
			tmp = p;
			p->next = p->next->next;
			xfree(tmp);
		}
}

void xsend(int cmd, const char *fmt, ...)
{
	struct packet *p;
	va_list ap;
	
	if (client_sock == -1)
		return;
	
	va_start(ap, fmt);
	p = packet_vprintf(fmt, ap);
	if (packet_send(client_sock, cmd, p)) {
		xclose(client_sock);
		client_sock = -1;
		prt(0, "\e<info> connection to client lost");
	}
	xfree(p);
}

static int do_auth(int fd)
{
	char junk[128], *x;
	int i;

	arc4random_stir();
	for (i = 0; i < 32; i++)
		junk[i] = arc4random_byte() % 90 + ' ';
	write(fd, junk, 32);
	
	strcpy(junk + 32, listen_password);
	x = count_md5(junk);
	
	memset(junk, 0, 32);
	read(fd, junk, 32);
	i = memcmp(junk, x, 32);
	
	xfree(x);
	
	if (i == 0)
		return 0;
	else
		return -1;
}

static void init_client_connection()
{
	sock_link(0, client_sock, client_handler, 0);
	
	send_colors();
	send_bools();
	win_sync(100);
	prt(0, "\e<info> init ok");
	eval("cc_client_connected();");
}

static void got_connection()
{
	int fd;
	size_t s;
	char *s1;
	
	union {
	struct sockaddr_in in4;
#ifdef IPv6
	struct sockaddr_in6 in6;
#endif
	} sin;
	s = sizeof(sin);
	fd = accept(listen_sock, (struct sockaddr*)&sin, &s);
	if (fd == -1)
		return;
	s1 = show_addr((struct sockaddr*)&sin, s);
	prtf(0, "\e<info> client connection from %s", s1);
	xfree(s1);

	alarm(10);
	if (do_auth(fd)) {
		close(fd);
		prt(0, "\e<warn> client connection failed !");
		alarm(0);
		return;
	}
	alarm(0);
	prt(0, "\e<info> client authenthicated");
	if (client_sock != -1)
		xclose(client_sock);
	client_sock = fd;
	write(client_sock, "ok", 2);
	init_client_connection();
}

static int listen_handler(int fd, void *data)
{
	(void)fd;
	(void)data;
	got_connection();
	return 0;
}

static int hist_pos, hist_size, hist_ptr;
static char **history;

void do_history(const char *win, const char *key, int *pos, char **cap)
{
	int copy = -1;
	
	(void)win;

	if (strcmp(key, "up") == 0 && hist_ptr > 0)
		copy = --hist_ptr;
	else if (strcmp(key, "down") == 0 && hist_ptr < hist_pos - 1)
		copy = ++hist_ptr;
	else if (strcmp(key, "down") == 0) {
		xfree(*cap);
		*cap = xstrdup("");
		*pos = 0;
		hist_ptr = hist_pos;
	}

	if (copy == -1) { /* check binings */ }
	else {
		xfree(*cap);
		*cap = xstrdup(history[copy]);
		*pos = strlen(*cap);
	}
}

void add_history(const char *win, const char *str)
{
	const char *p;
	
	(void)win;
	p = str;
	while (*p == ' ')
		p++;
	if (*p == 0)
		return;		/* skip empty lines */
	
	if (++hist_pos > hist_size) 
		history = (char**)xrealloc(history, 
				(hist_size += 20) * sizeof(char*));
	history[hist_pos - 1] = xstrdup(str);
	hist_ptr = hist_pos;
}

static int client_handler(int fd, void *data)
{
	struct packet *p;
	char *arg, *win;
	int cmd;

	(void)data;
	
	p = packet_recv(fd, &cmd);
	if (p == 0) {
		client_sock = -1;
		return -1;
	}
		
	switch (cmd) {
	case CMD_INPUT:
		packet_scanf(p, "%s%s", &win, &arg);
		do_input(win, arg);
		add_history(win, arg);
		xfree(arg);
		xfree(win);
		break;
	case CMD_KEY:
		{
			char *key, *cap;
			int pos;
			
			packet_scanf(p, "%s%s%d%s", &win, &key, &pos, &cap);
			do_history(win, key, &pos, &cap);
			xsend(CMD_SET_INPUT, "%s%d%s", win, pos, cap);
			xfree(key);
			xfree(win);
			xfree(cap);
		}
		break;
	}

	xfree(p);
	
	return 0;
}

void sock_link(int beg, int fd, int (*handler) (int fd, void *data), 
	       void *data)
{
	struct socket *s, *p;
	
	s = NEW(struct socket);
	
	if (beg || sock_head == 0) {
		s->next = sock_head;
		sock_head = s;
	} else {
		s->next = 0;
		for (p = sock_head; p->next; p = p->next)
			/* nothing */;
		p->next = s;
	}
		
	s->fd = fd;
	s->data = data;
	s->handler = handler;
}

void link_listen_socket()
{
	sock_link(1, listen_sock, listen_handler, 0);
}

static u32 bool1;
static char colors[100];

void send_colors()
{
#define getcolor(C_XXX, xxx) \
	colors[C_XXX] = hv("colors", xxx)->var_int
	
	memset(colors, 0, 100);
	
	getcolor(C_STATUS_NORMAL, "status_normal");
	getcolor(C_STATUS_CHANGED, "status_changed");
	getcolor(C_STATUS_ACTIVE, "status_active");
	getcolor(C_MENU, "menu");
	getcolor(C_MENU_SELECTED, "menu_selected");
	getcolor(C_MENU_ALT, "menu_alt");
	getcolor(C_MENU_ALT_SELECTED, "menu_alt_selected");
	
#undef getcolor
	xsend(CMD_CONF_COLORS, "%s", colors);
}

void send_bools()
{
#define getbool(B_XXX, xxx) \
	bool1 |= hv("bools", xxx)->var_int ? (1 << B_XXX) : 0

	bool1 = 0;
	
	getbool(B_CONTRAST_COLORS, "contrast_colors"); 
	getbool(B_COPY_STATUS, "copy_status"); 
	getbool(B_SHORT_STATUS_LINE, "short_status_line"); 
	getbool(B_NO_COLORS, "no_colors"); 
	getbool(B_NO_CONT, "no_cont"); 
	getbool(B_HALF_SCROLL, "half_scroll"); 
	getbool(B_SCROLL_ON_PRINT, "scroll_on_print"); 
	getbool(B_WINDOW_INPUT_LINE, "window_input_line"); 
	getbool(B_INVISIBLE, "invisible"); 
#undef getbool
	xsend(CMD_CONF_BOOL1, "%d", bool1);
}
