#include "config.h"
#include "lib.h"
#include "proto.h"

#include <stdarg.h>
#include <unistd.h>
#include <netinet/in.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

static int xread(int fd, void *buf, int size)
{
	int n;
	while (size) {
		n = read(fd, buf, size);
		if (n <= 0)
			return -1;
		size -= n;
		((char *) buf) += n;
	}
	return 0;
}

static int xwrite(int fd, const void *buf, int size)
{
	int n;
	while (size) {
		n = write(fd, buf, size);
		if (n == -1)
			return -1;
		size -= n;
		((const char *) buf) += n;
	}
	return 0;
}

struct packet *packet_recv(int fd, int *cmd)
{
	struct packet *r, *r2;
	u32 s;

	if (xread(fd, &s, 4))
		return 0;
	s = ntohl(s);
	if (s > 0xfffff)
		return 0;	/*oops */
	r = (struct packet *)xmalloc(s + 4);
	r->size = s;
	if (xread(fd, r->data, s)) {
		xfree(r);
		return 0;
	}
	packet_scanf(r, "%d%p", cmd, &r2);
	xfree(r);
	return r2;
}

int packet_send(int fd, int cmd, const struct packet *p)
{
	struct packet *s;
	int r;

	s = packet_printf("%d%p", cmd, p);
	r = xwrite(fd, s, ntohl(s->size) + 4);
	xfree(s);
	
	return r;
}

static char *do_printf(const char *p, char *d, va_list ap, int real)
{
	char *tmp1;
	int tmp2;
	struct packet *tmp;
	
	for (;;) {
		while (*p && *p != '%')
			p++;
		if (*p)
			p++;
		else
			break;
			
		switch (*p) {
		case 's':
			tmp1 = va_arg(ap, char *);
			if (real)
				strcpy(d, tmp1);
			d += strlen(tmp1) + 1;
			break;
		case 'd':
			tmp2 = va_arg(ap, u32);
			if (real)
				*((u32 *) d) = htonl(tmp2);
			d += 4;
			break;
		case 'p':
			tmp = va_arg(ap, struct packet *);
			if (real)
				memcpy(d, tmp, ntohl(tmp->size) + 4);
			d += ntohl(tmp->size) + 4;
			break;
		}
	}
	va_end(ap);

	return d;
}

struct packet *packet_vprintf(const char *fmt, va_list ap)
{
	int len = 0;
	struct packet *r;

	len = (int)do_printf(fmt, 0, ap, 0);
	
	r = (struct packet*)xmalloc(4 + len);
	r->size = htonl(len);

	do_printf(fmt, r->data, ap, 1);
	
	return r;
}

struct packet *packet_printf(const char *fmt, ...)
{
	va_list ap;

	va_start(ap, fmt);
	return packet_vprintf(fmt, ap);
}

void packet_scanf(const struct packet *x, const char *p, ...)
{
	const char *s;
	struct packet *tmp;
	int len;
	va_list ap;
	
	s = x->data;
	va_start(ap, p);

	for (;;) {
		while (*p && *p != '%')
			p++;
		if (*p)
			p++;
		else
			break;

		switch (*p) {
		case 's':
			*(va_arg(ap, char **)) = xstrdup(s);
			s = strchr(s, 0) + 1;
			break;
		case 'd':
			*(va_arg(ap, int *)) = ntohl(*((u32 *) s));
			s += 4;
			break;
		case 'p':
			len = ntohl(((struct packet *) s)->size);
			s += 4;
			tmp = (struct packet *)xmalloc(len + 4);
			memcpy(tmp->data, s, len);
			s += len;
			tmp->size = len;
			*(va_arg(ap, struct packet **)) = tmp;
			break;
		}
	}
	if ((char*)s - (char*)x->data != (int)x->size)
		abort();
	va_end(ap);
}
