#ifndef _PROTO_H
#define _PROTO_H 1

#include <stdarg.h>

struct packet {
	u32 size;
	byte data[0];
};

/* packet: %d%p : command, packet */

enum {
CMD_CROAK = 0,		/* error */
/* commands sent from bot to iface */
CMD_PRINT = 0x10000,	/* fmt: %s%s : win_name, string */
CMD_EXIT,		/* fmt: %s : reason */
CMD_NEW_WINDOW,		/* fmt: %s : win_name */
CMD_KILL_WINDOW,	/* fmt: %s : win_name */
CMD_SET_INPUT,		/* fmt: %s%d%s : win_name, pos, caption */
CMD_SET_STATUS_LINE,	/* fmt: %s : status line */
/* config commands */
CMD_CONF_COLORS,	/* fmt: %s : colors */
CMD_CONF_BOOL1,		/* fmt: %d : bool1 */

/* commands sent the other way around */
CMD_INPUT = 0x20000,	/* fmt: %s%s : win_name, string */
CMD_KEY,		/* fmt: %s%s%d%s : win_name, key_name, 
			 *   pos_in_input_line, input_line */ 
CMD_FREEZE,		/* fmt: %d : do_freeze */
};

/* f for packets: 
 * %s nul terminated string 
 * %p another packet
 * %d 4-byte int, big endian 
 */
struct packet *packet_printf(const char *fmt, ... )
	__attribute__ ((format (printf, 1, 2)));
void packet_scanf(const struct packet *p, const char *fmt, ... );
int packet_send(int fd, int cmd, const struct packet *p);
struct packet *packet_recv(int fd, int *cmd);
struct packet *packet_vprintf(const char *fmt, va_list ap);

/* --------------------------- config ---------------------------- */

#define COLOR(fg,bg) (fg|(bg<<4))
enum COLORS { 
	BLACK, BLUE, GREEN, CYAN, RED, MAGENTA, BROWN, LIGHTGRAY, 
	DARKGRAY, LIGHTBLUE, LIGHTGREEN, LIGHTCYAN, LIGHTRED,
	LIGHTMAGENTA, YELLOW, WHITE
};

/* bool configs */
enum {
B_CONTRAST_COLORS,	/* use contrast colors */
B_COPY_STATUS,		/* copy everything from status window to current */
B_SHORT_STATUS_LINE,	/* is status line with windows to be compressed */
B_NO_COLORS,		/* don't use colors */
B_HALF_SCROLL,		/* scroll by half of screen (like bx) */
B_SCROLL_ON_PRINT,	/* scroll window down when sth is printed */
B_WINDOW_INPUT_LINE,	/* each win has own input line */
/* for use by script */
B_INVISIBLE,		/* inivisible input line for entering passwds */
B_NO_CONT,		/* do not use "     " continue string */
};
/* color settings */
enum {
C_STATUS_NORMAL,
C_STATUS_CHANGED,
C_STATUS_ACTIVE,
C_MENU,
C_MENU_SELECTED,
C_MENU_ALT,
C_MENU_ALT_SELECTED,
};
int get_bool(int no);
int get_color(int no);

#endif /* _PROTO_H */

