/* include gnu.gpl, (*) 2000 Micha 'malekith' Moskal <malekith@topnet.pl>. 
 * $Id: keyword.c,v 1.12 2000/08/06 12:05:35 malekith Exp $ */

#include "parser.h"
#include "var.h"
#include "proc.h"

struct keyword {
	const char *word;
	token *(*handle)(CTX *ctx, struct keyword *self);
};

extern struct keyword keywords[]; 

static struct keyword *find_kw(char *word)
{
	int i;

	for (i = 0; keywords[i].word; i++)
		if (strcmp(keywords[i].word, word) == 0)
			break;
			
	if (keywords[i].word)
		return keywords + i;
	else
		return 0;
}

int is_keyword(char *s)
{
	return !!find_kw(s);
}

token *handle_keyword(CTX * ctx, char *word)
{
	struct keyword *k;
	k = find_kw(word);
	if (!k)
		abort();
	return k->handle(ctx, k);
}

static token *kw_hash(CTX * ctx, struct keyword *self)
{
	token *head = 0, *p = 0, *x;
	
	for (;;) {
		x = get_expression(ctx);
		if (x == 0)
			break;
		if (x->id == ID_VAR) {	/* only have to rename */
			/* and link corz */
			link_end(&head,&p,x);
		} else {
			poops(ctx, "bad var definition");
			break;
		}
		
		p->id = *self->word == 'i' ? ID_DEF_IHASH : 
			*self->word == 'p' ? ID_DEF_PHASH : ID_DEF_SHASH;

		if (ctx->type != T_OP)
			poops(ctx, "non OP found");
		else if (ctx->iv == ';')
			return head;
		else if (ctx->iv == ',')
			continue;
			
		poops(ctx, "wrong OP found");
		break;
	}
	kill_token(head);
	return 0;
}

static token *kw_int(CTX * ctx, struct keyword *self)
{
	token *head = 0, *p = 0, *x;
	
	for (;;) {
		x = get_expression(ctx);
		if (x == 0)
			break;
		if (x->id == ID_EXPR2 && x->u.number == 0x100) {	
			/* assign, we're not intrested in x itself, buf in 
			 * x->right (a var) and x->left a value */
			x->left->left = x->right; 
			/* ^^^^^^^^ left shall contain caption */
			x->right = 0;
			link_end(&head,&p,x->left);
			x->left = 0;
			kill_token(x);
		} else if (x->id == ID_VAR) {	/* only have to rename */
			/* and link corz */
			link_end(&head,&p,x);
		} else {
			poops(ctx, "bad var definition");
			break;
		}
		
		p->id = *self->word == 'i' ? ID_DEF_INT : 
			*self->word == 's' ? ID_DEF_STRING : ID_DEF_PTR;

		if (ctx->type != T_OP)
			poops(ctx, "non OP found");
		else if (ctx->iv == ';')
			return head;
		else if (ctx->iv == ',')
			continue;
			
		poops(ctx, "wrong OP found");
		break;
	}
	kill_token(head);
	return 0;
}

static token *kw_else(CTX *ctx, struct keyword *self)
{
	(void)self;

	poops(ctx, "`else' w/o `if'");
	return 0;
}

static token *kw_return(CTX * ctx, struct keyword *self)
{
	token *x;

	x = make_token(ctx,
		       *self->word == 'r' ? ID_RETURN :
		       *self->word == 'b' ? ID_BREAK : ID_CONTINUE, 0, 0);
		       
	get_token(ctx);
	
	if (ctx->type == T_OP && ctx->iv == ';')
		x->left = 0;
	else {
		ctx->leave_it = 1;
		if ((x->left = get_expression(ctx)) == 0) {
			kill_token(x);
			return 0;
		}
	}
	
	if (ctx->type != T_OP || ctx->iv != ';') {
		poops(ctx, "%s: bad end", self->word);
		kill_token(x);
		return 0;
	}
	
	return x;
}

static token *kw_void(CTX *ctx, struct keyword *self)
{
	(void)self;
	poops(ctx, "void keyword used");
	return 0;
}

static token *xget_bool(CTX * ctx, const char *name)
{
	token *e;

	if (get_token(ctx) == 0) {
		poops(ctx, "%s: EOF", name);
		return 0;
	}
	if (ctx->type != T_OP || ctx->iv != '(') {
		poops(ctx, "%s: no (", name);
		return 0;
	}
	if ((e = get_expression(ctx)) == 0)
		return 0;
	if (ctx->type != T_OP || ctx->iv != ')') {
		poops(ctx, "%s: no )", name);
		kill_token(e);
		return 0;
	}
	return e;
}

static token *kw_dowhile(CTX *ctx, struct keyword *self)
{
	token *e = 0, *b = 0, *t;
	
	(void)self;

	if ((b = get_instruction(ctx)) == 0)
		return 0; 

	get_token(ctx);
	
	if (ctx->type != T_WORD || strcmp(ctx->sv, "while")) {
		kill_token(b);
		poops(ctx, "``do'' not followed with ``while''");
		return 0;
	}
	
	if ((e = xget_bool(ctx, "do_while")) == 0) {
	    	kill_token(b);
		return 0;
	}
	
	t = make_token(ctx, ID_DO_WHILE, 0, 0);
	t->u.top = e;
	t->left = b;
	return t;
}

static token *kw_while(CTX *ctx, struct keyword *self)
{
	token *e, *b, *t;
	
	(void)self;

	if ((e = xget_bool(ctx, "while")) == 0)
		return 0;
	if ((b = get_instruction(ctx)) == 0) {
		kill_token(e);
		return 0;
	}
	t = make_token(ctx, ID_WHILE, 0, 0);
	t->u.top = e;
	t->left = b;
	return t;
}

static token *kw_for(CTX *ctx, struct keyword *self)
{
	token *pre = 0, *cond = 0, *post = 0, *body = 0;
	token *t;

	(void)self;
	if (get_token(ctx) == 0) {
		poops(ctx, "for: EOF");
		return 0;
	}

	if ((pre = get_instruction(ctx)) == 0 ||
	    (cond = get_instruction(ctx)) == 0 ||
	    (post = get_expression(ctx)) == 0 ||
	    (body = get_instruction(ctx)) == 0) {
	    	kill_token(pre);
	    	kill_token(post);
	    	kill_token(cond);
	    	kill_token(body);
	    	return 0;
	}
	for (t = body; t->next; t = t->next)
		/* nothing */ ;
	t->next = post;
	for (t = pre; t->next; t = t->next)
		/* nothing */ ;
	t->next = make_token(ctx, ID_WHILE, 0, 0);
	t->next->u.top = cond;
	t->next->left = body;

	return pre;
}

static token *kw_if(CTX * ctx, struct keyword *self)
{
	token *e, *b, *el = 0, *x;

	(void) self;

	if ((e = xget_bool(ctx,"if")) == 0)
		return 0;
	
	if ((b = get_instruction(ctx)) == 0) {
		kill_token(e);
		return 0;
	}
	
	if (get_token(ctx) && ctx->type == T_WORD
	    && strcmp(ctx->sv, "else") == 0) {
		el = get_instruction(ctx);
		if (el == 0) {
			kill_token(e);
			kill_token(b);
			return 0;
		}
	} else
		ctx->leave_it = 1;

	x = make_token(ctx, ID_IF, 0, 0);
	x->u.top = e;
	x->left = b;
	x->right = el;
	return x;
}

static token *make_arg_list(CTX * ctx)
{
	token *h = 0, *x = 0, *pp;
	char *type, *name;
	int tt = V_INT;

	for (;;) {
		get_token(ctx);
		
		if (ctx->type == 0 || 
		    (ctx->type == T_OP && ctx->iv == ')'))
		    	return 0;
			
		if (ctx->type != T_WORD) {
			poops(ctx, "arg list: no word found");
			break;
		}
		type = ctx->sv;
		ctx->sv = 0;
		get_token(ctx);
		if (ctx->type == T_WORD) {
			if (strcmp(type, "int") == 0)
				tt = V_INT;
			else if (strcmp(type, "string") == 0)
				tt = V_STRING;
			else {
				poops(ctx, "arg list: wrog type '%s'",
				      type);
				xfree(type);
				break;
			}
			xfree(type);
			name = ctx->sv;
			ctx->sv = 0;
			get_token(ctx);
		} else
			name = type;
		pp =
		    make_token(ctx,
			       tt == V_INT ? ID_DEF_INT : ID_DEF_STRING, name,
			       0);
		link_end(&h,&x,pp);
		
		if (ctx->type == 0)
			return h; /* for cprocs */
			
		if (ctx->type != T_OP || (ctx->iv != ',' && ctx->iv != ')')) {
			poops(ctx, "arg list: no/wrong OP");
			break;
		}
		if (ctx->iv == ')')
			return h;
	}
	kill_token(h);
	return (token*)-1;
}

/* for cprocs */
byte *bc_make_arg_list(const char *s)
{
	char *tmp;
	CTX ctx;
	token *t;
	
	tmp = xstrdup(s);
	memset(&ctx, 0, sizeof(ctx));
	ctx.p = tmp;
	ctx.fn = "CPROC ARG LIST";
	t = make_arg_list(&ctx);
	xfree(tmp);
	if (t == (token *) -1)
		abort();
		
	return pack(t, 0, 0, 0);
}

static token *kw_walk(CTX * ctx, struct keyword *self)
{
	token *x;
	char *s;
	
	(void) self;
	
	get_token(ctx);
	if (ctx->type != T_WORD) {
		poops(ctx, "walk: not follwed by a word");
		return 0;
	}
	s = ctx->sv;
	ctx->sv = 0;
	x = make_token(ctx, ID_WALK, s, 0);
	
	if ((x->left = get_instruction(ctx)) == 0) {
		kill_token(x);
		return 0;
	}

	return x;
}
extern int alloc_print;
void print_x(token*);

static token *kw_proc(CTX * ctx, struct keyword *self)
{
	char *s;
	token *body, *args;
	struct proc *p;

	(void) self;
	get_token(ctx);
	if (ctx->type != T_WORD) {
		poops(ctx, "proc: not follwed by a word");
		return 0;
	}
	s = ctx->sv;
	ctx->sv = 0;

	get_token(ctx);
	if (ctx->type != T_OP || ctx->iv != '(') {
		poops(ctx, "proc: no (");
		xfree(s);
		return 0;
	}

	if ((args = make_arg_list(ctx)) == (token *) -1) {
		xfree(s);
		return 0;
	}
	if ((body = get_instruction(ctx)) == 0) {
		kill_token(args);
		xfree(s);
		return 0;
	}

	if (getenv("DUMP_TREE")) {
		token *t, *tmp;
		
		printf("proc %s(", s);
		for (t = args; t; t = t->next) {
			tmp = t->next;
			t->next = 0;
			print_x(t);
			if ((t->next = tmp))
				printf(", ");
		}
		printf(")\n{\n");
		print_x(body);
		printf(";\n}\n");
	}

	p = make_proc(args, body, s);
	xfree(s);
	add_proc(ctx->frm, p);

	return make_token(ctx, ID_INT, 0, 1);	/* dummy true value */
}

struct keyword keywords[] = {
	{ "int", kw_int },
	{ "string", kw_int },
	{ "ptr", kw_int },
	{ "shash", kw_hash },
	{ "hash", kw_hash },
	{ "ihash", kw_hash },
	{ "phash", kw_hash },
	{ "break", kw_return },
	{ "else", kw_else },
	{ "return", kw_return },
	{ "continue", kw_return },
	{ "while", kw_while },
	{ "void", kw_void },
	{ "if", kw_if },
	{ "proc", kw_proc },
	{ "for", kw_for },
	{ "do", kw_dowhile },
	{ "walk", kw_walk },
	{ 0, 0 }
};
