/* include gnu.gpl, (*) 2000 Micha 'malekith' Moskal <malekith@topnet.pl>. 
 * $Id: regexp.c,v 1.6 2000/08/04 21:28:14 malekith Exp $ */

#include "vm.h"
#include "parser.h"

#include "pcre.h"


struct regexp_info {
	int no;			/* number of ()'s in RE */
	int real;		/* number of ints allocated at m */
	int matched;		/* number of matched ()'s in RE */
	int *m;			/* ptr to ()'s */
	char *string;		/* matched string */
};

var *regexp_get(vm_token *t, frame *f, int no)
{
	var *r = 0;

	if (f->regexp == 0) 
		croak(t, "frame doesn't have regexp info "
			"(using $%d etc. w/o prevoius match)", no);
	else if (f->regexp->no < no)
		croak(t, "last regexp has only %d ()'s (request for $%d)",
			f->regexp->no, no);
	else if (!f->regexp->matched)
		/* croak(t, "last RE didn't match"); */
		r = new_sv(0, "");
	else {
		int s, e;

		e = f->regexp->m[no * 2 + 1];
		s = f->regexp->m[no * 2];
		
		e -= s;
		r = new_svn(0, e);
		memcpy(r->var_string, f->regexp->string + s, e);
		r->var_string[e] = 0;
	}
	
	return r;
}

int regexp_do(const char * x1, byte *x2, frame * f)
{
	vm_token t;
	int needed;
	
	unpack(&t, x2);
	
	if (f->regexp == 0) {
		f->regexp = (struct regexp_info *)
		    		xmalloc(sizeof(struct regexp_info));
		f->regexp->real = 0;
		f->regexp->m = 0;
		f->regexp->string = 0;
	}

	if (pcre_fullinfo(t.data_ptr, 0, PCRE_INFO_CAPTURECOUNT, 
			  &f->regexp->no))
		abort();
		
	needed = (f->regexp->no + 1) * 3;
	
	if (f->regexp->real < needed) {
		xfree(f->regexp->m);
		f->regexp->real = needed;
		f->regexp->m = (int*)xmalloc(sizeof(int) * f->regexp->real);
	}
	
	xfree(f->regexp->string);
	f->regexp->string = xstrdup(x1);

	f->regexp->matched = pcre_exec(t.data_ptr, 0, f->regexp->string,
					strlen(f->regexp->string), 0,
					0, f->regexp->m, needed);
					
	if (f->regexp->matched == PCRE_ERROR_NOMATCH)
		f->regexp->matched = 0;
	else if (f->regexp->matched <= 0)
		abort();
		
	return f->regexp->matched;
}

void regexp_info_free(struct regexp_info *r)
{
	if (r == 0)
		return;
	xfree(r->m);
	xfree(r->string);
	xfree(r);
}

const char *regexp_make(const char *string, token *t)
{
	const char *err;
	int eoff;
	size_t s;

	t->data_ptr = pcre_compile(string + 1, 
			((*string & 0x2) ? PCRE_CASELESS : 0) |
			((*string & 0x4) ? PCRE_MULTILINE : 0) |
			((*string & 0x8) ? PCRE_EXTENDED : 0),
			&err, &eoff, 0);

	if (t->data_ptr == 0)
		return err ? err : "unknown regexp error";

	if (pcre_fullinfo(t->data_ptr, 0, PCRE_INFO_SIZE, &s))
		abort();
		
	t->data_len = (int)s;
	
	return 0;
}
