/* include gnu.gpl, (*) 2000 Micha 'malekith' Moskal <malekith@topnet.pl>. 
 * $Id: inputline.cc,v 1.4 2000/08/27 18:32:51 malekith Exp $ */

#include "sniffy.h"
#include <string.h>
#include <stdio.h>

InputLine *input_line;

int menu_active()
{
	return 0;
}

int menu_handle_key(int k)
{
	return k;
}

int vstrlen(const char *p)
{
	int len = 0;
	while (*p) {
		if (*p == 27 && p[1]) p += 2;
		else { p++; len++; }
	}
	return len;
}

void InputLine::draw()
{
	char *p;

	if (scr_offset > (int)strlen(caption))
		scr_offset = strlen(caption) - 20;
	if (scr_offset < 0)
		scr_offset = 0;
		
	if (get_bool(B_INVISIBLE))
		p = xstrdup(prompt());
	else
		p = xf("%s%s", prompt(), caption + scr_offset);
		
	put_line(0,s_height-1,p);
	xfree(p);
	go(curs_pos + scr_offset + vstrlen(prompt()), s_height - 1);
}

void InputLine::handle_input()
{
	int k = getch();
	
	while (k) {
		k = menu_handle_key(k);
		if (k)
			handle_key(k);
		k = getch();
	}
}

void InputLine::fix_pos()
{
	if (curs_pos < 0) curs_pos = 0;
	if (curs_pos > (int)strlen(caption)) curs_pos = strlen(caption);
	
	int w = strlen(caption) - (s_width - strlen(prompt()));
	if (w < 0) { scr_offset = 0; return; }
	if (scr_offset > curs_pos) scr_offset = curs_pos;
	if (scr_offset < w) scr_offset = w;
}

void switch_to_win(int k)
{
	char *p = "1234567890!@#$%^&*()";
	int n = strchr(p, k) - p + 1;
	if (n < 1 || n > 20)
		return;
	if (!win20[n])
		return;
	select_win(win20[n]);
}

void input_key(const char *key, char *caption, int *pos, int clen);

void InputLine::handle_key(int k)
{
	char *x = 0;
	char buf[10];

	switch (k) {
	case '\n':
	case '\r':	x = "enter"; 	break;
	case key_up:	x = "up";	break;
	case key_down:	x = "down";	break;
	case '\t':	x = "tab";	break;
	case key_right: curs_pos++;	break;
	case key_left:	curs_pos--;	break;
	case key_home:	curs_pos = 0;	break;
	case key_end:
		curs_pos = strlen(caption);
		break;
	case '\b':
		if (--curs_pos < 0) break;
		/* FALL THRU */
	case key_del:
		{
			char *p = caption + curs_pos;
			if (*p)
				memmove(p, p+1, strlen(p));
		}
		break;
	case key_ctrl('a'):
	case key_ctrl('c'):
	case key_ctrl('b'):
	case key_ctrl('v'):
	case key_ctrl('o'):
	case key_ctrl('_'):
		self_insert(k);
		break;
	case key_ctrl('e'):
	case key_pgup: win_topmost->pgup();	break;
	case key_ctrl('d'):
	case key_pgdn: win_topmost->pgdn(); 	break;
	case key_ctrl('l'): f_repaint();	break;
	case key_f(6):
	case key_ctrl('w'):
		select_win(win_topmost->next ? : win_head);
		break;
	case 'x' | key_meta:
		kill_curses();
		exit(1);
	default:
		if ((k & (32|64)) == 0) {
			sprintf(buf,"c-%c", k|32);
			x = buf;
		} else if (k & key_meta) {
			k &= 0xff;
			if ((k >= '0' && k <= '9') || strchr("!@#$%^&*()",k))
				switch_to_win(k);
			else {
				sprintf(buf,"m-%c", k);
				x = buf;
			}
		} else if (k > key_f(0) && k <= key_f(30)) {
			sprintf(buf,"f-%d",k - key_f(0));
			x = buf;
		} else
			self_insert(k);
		break;
	}

	fix_pos();

	if (x) {
		input_key(x, caption, &curs_pos, sizeof(caption) - 1);
		fix_pos();
	}
}

void InputLine::self_insert(int k)
{
	if (curs_pos > (int)sizeof(caption)-2)
		return;
	char *p = caption+curs_pos;
	memmove(p+1,p,strlen(p)+1);
	*p = k;
	curs_pos++;
}
